from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from fastapi import status

from ohc_backend.services.ha_service import HomeAssistantService


class TestHomeAssistantService:
    """Tests for HomeAssistantService."""

    @pytest.fixture
    def ha_service(self):
        """Create a test instance of HomeAssistantService with patched session."""
        with patch("aiohttp.ClientSession"):
            service = HomeAssistantService(
                "http://localhost:8123", "test_token")
        return service

    @pytest.mark.asyncio
    async def test_make_request_success(self, ha_service):
        """Test successful API requests."""
        # Define the expected return value
        mock_data = {"key": "value"}

        # Create a complete patch for the request method and context manager
        with patch.object(ha_service, "session") as mock_session:
            # Create a mock response that will be returned by the context manager
            mock_response = MagicMock()
            mock_response.status = status.HTTP_200_OK
            mock_response.ok = True
            mock_response.json = AsyncMock(return_value=mock_data)

            # Configure the async context manager
            mock_cm = AsyncMock()
            mock_cm.__aenter__.return_value = mock_response
            mock_session.request.return_value = mock_cm

            # Call the method
            result = await ha_service.make_request("GET", "http://test/endpoint")

            # Verify results
            assert result == mock_data
            mock_session.request.assert_called_once_with(
                "GET", "http://test/endpoint")

    async def test_close(self, ha_service):
        """Test closing the service."""
        # Create a proper mock session
        mock_session = MagicMock()
        mock_session.closed = False
        mock_session.close = AsyncMock()

        # Replace the service's session with our mock
        ha_service.session = mock_session

        # Call the method
        await ha_service.close()

        # Verify the result
        mock_session.close.assert_called_once()

    @pytest.mark.asyncio
    async def test_json_to_yaml(self, ha_service):
        """Test YAML conversion with dict input."""
        dict_input = {"name": "Test", "value": 123}
        yaml_result = ha_service.json_to_yaml(dict_input)

        expected_yaml = "name: Test\nvalue: 123\n"
        assert yaml_result == expected_yaml

    @pytest.mark.asyncio
    async def test_get_automation_content(self, ha_service):
        """Test getting automation content."""
        automation_dict = {"id": "test_auto", "name": "Test Automation"}

        # Patch both methods we'll use
        with patch.object(ha_service, "make_request",
                          new_callable=AsyncMock,
                          return_value=automation_dict) as mock_request, \
            patch.object(ha_service, "json_to_yaml",
                         return_value="name: Test Automation\n") as mock_yaml:

            result = await ha_service.get_automation_content("test_auto")

            mock_request.assert_called_with(
                "GET",
                "http://localhost:8123/api/config/automation/config/test_auto"
            )
            mock_yaml.assert_called_with(automation_dict)
            assert result == "name: Test Automation\n"

    @pytest.mark.asyncio
    async def test_make_request_not_found(self, ha_service):
        """Test 404 response handling."""
        with patch.object(ha_service, "session") as mock_session:
            # Create a mock response
            mock_response = MagicMock()
            mock_response.status = status.HTTP_404_NOT_FOUND

            # Configure the async context manager
            mock_cm = AsyncMock()
            mock_cm.__aenter__.return_value = mock_response
            mock_session.request.return_value = mock_cm

            # Call the method
            result = await ha_service.make_request("GET", "http://test/endpoint")

            # Verify results
            assert result is None
            mock_session.request.assert_called_once()

    @pytest.mark.asyncio
    async def test_make_request_error(self, ha_service):
        """Test error response handling."""
        with patch.object(ha_service, "session") as mock_session:
            # Create a mock response
            mock_response = MagicMock()
            mock_response.status = status.HTTP_500_INTERNAL_SERVER_ERROR
            mock_response.ok = False
            mock_response.json = AsyncMock(
                return_value={"error": "server error"})
            mock_response.raise_for_status = MagicMock(
                side_effect=Exception("HTTP Error"))

            # Configure the async context manager
            mock_cm = AsyncMock()
            mock_cm.__aenter__.return_value = mock_response
            mock_session.request.return_value = mock_cm

            # Call the method and check for exception
            with pytest.raises(Exception):
                await ha_service.make_request("GET", "http://test/endpoint")

    @pytest.mark.asyncio
    async def test_get_script_content(self, ha_service):
        """Test getting script content."""
        script_dict = {"id": "test_script", "name": "Test Script"}

        with patch.object(ha_service, "make_request",
                          new_callable=AsyncMock,
                          return_value=script_dict) as mock_request, \
            patch.object(ha_service, "json_to_yaml",
                         return_value="name: Test Script\n") as mock_yaml:

            result = await ha_service.get_script_content("script.test_script")

            mock_request.assert_called_with(
                "GET",
                "http://localhost:8123/api/config/script/config/test_script"
            )
            mock_yaml.assert_called_with(script_dict)
            assert result == "name: Test Script\n"

    @pytest.mark.asyncio
    async def test_get_automation(self, ha_service):
        """Test getting an automation entity."""
        from ohc_backend.models.ha_entity import Automation

        mock_state = {
            "entity_id": "automation.test_auto",
            "state": "on",
            "attributes": {
                "id": "test_auto",
                "friendly_name": "Test Automation"
            },
            "last_changed": "2023-01-01T00:00:00Z"
        }

        with patch.object(ha_service, "make_request",
                          new_callable=AsyncMock,
                          return_value=mock_state):

            result = await ha_service.get_automation("test_auto")

            assert isinstance(result, Automation)
            assert result.entity_id == "automation.test_auto"
            assert result.state == "on"
            assert result.friendly_name == "Test Automation"
            assert result.automation_id == "test_auto"

    @pytest.mark.asyncio
    async def test_get_script(self, ha_service):
        """Test getting a script entity."""
        from ohc_backend.models.ha_entity import Script

        mock_state = {
            "entity_id": "script.test_script",
            "state": "off",
            "attributes": {
                "friendly_name": "Test Script"
            },
            "last_changed": "2023-01-01T00:00:00Z"
        }

        with patch.object(ha_service, "make_request",
                          new_callable=AsyncMock,
                          return_value=mock_state):

            result = await ha_service.get_script("test_script")

            assert isinstance(result, Script)
            assert result.entity_id == "script.test_script"
            assert result.state == "off"
            assert result.friendly_name == "Test Script"

    @pytest.mark.asyncio
    async def test_get_all_automations_and_scripts(self, ha_service):
        """Test getting all automations and scripts."""
        from ohc_backend.models.ha_entity import Automation, Script

        mock_states = [
            {
                "entity_id": "automation.test_auto",
                "state": "on",
                "attributes": {
                    "id": "test_auto",
                    "friendly_name": "Test Automation"
                },
                "last_changed": "2023-01-01T00:00:00Z"
            },
            {
                "entity_id": "script.test_script",
                "state": "off",
                "attributes": {
                    "friendly_name": "Test Script"
                },
                "last_changed": "2023-01-01T00:00:00Z"
            },
            {
                "entity_id": "light.living_room",  # Should be filtered out
                "state": "on",
                "attributes": {},
                "last_changed": "2023-01-01T00:00:00Z"
            }
        ]

        with patch.object(ha_service, "make_request",
                          new_callable=AsyncMock,
                          return_value=mock_states):

            results = await ha_service.get_all_automations_and_scripts()

            assert len(results) == 2
            assert isinstance(results[0], Automation)
            assert isinstance(results[1], Script)
            assert results[0].entity_id == "automation.test_auto"
            assert results[1].entity_id == "script.test_script"



"""GitHub client that orchestrates all operations."""

import logging

from .auth import GitHubAuthManager
from .content import GitHubContentManager
from .models import GithubRepositoryRequestConfig
from .repository import GitHubRepositoryManager, RepositoryInitializer
from .rest import GitHubRestAPI

logger = logging.getLogger(__name__)


class GitHubClient:
    """High-level GitHub client that orchestrates all operations."""

    def __init__(self, api_url: str, client_id: str, access_token: str | None) -> None:
        """Initialize the GitHub client."""
        self.rest_api = GitHubRestAPI(api_url)
        self.auth_manager = GitHubAuthManager(client_id)
        self.repo_manager = GitHubRepositoryManager(self.rest_api)
        self.repo_initializer = RepositoryInitializer(self.rest_api)
        self._content_manager: GitHubContentManager | None = None

        if access_token:
            self.rest_api.set_auth_token(access_token)

    async def authenticate(self, scope: str) -> str:
        """Authenticate and configure the API client."""
        token = await self.auth_manager.authenticate(scope)
        self.rest_api.set_auth_token(token)
        return token

    async def init_repository(self, config: GithubRepositoryRequestConfig) -> None:
        """Initialize or connect to a repository."""
        repository = await self.repo_manager.find_repository(config.name)
        if not repository:
            logger.info("Repository not found, creating new repository")
            repository = await self.repo_manager.create_repository(config)
            await self.repo_initializer.initialize_with_readme(repository.full_name)

        self._content_manager = GitHubContentManager(
            self.rest_api, repository.full_name)

    @property
    def content(self) -> GitHubContentManager:
        """Access to content operations (requires initialized repository)."""
        if not self._content_manager:
            msg = "Repository not initialized. Call init_repository first."
            raise ValueError(msg)
        return self._content_manager

    async def close(self) -> None:
        """Close rest api session."""
        await self.rest_api.session.close()



"""GitHub authentication API client and manager."""
import asyncio
import logging
import time

import aiohttp

from .base import GitHubBaseAPI
from .errors import GitHubAuthError, GitHubError
from .models import DeviceFlowInfo, TokenResponse

logger = logging.getLogger(__name__)


class AuthenticationTimeoutError(GitHubAuthError):
    """Raised when authentication flow times out."""

    def __init__(self, timeout_minutes: int) -> None:
        """Initialize the error."""
        super().__init__(
            message=f"GitHub authentication timed out after {timeout_minutes} minutes",
            error_type="auth_timeout"
        )


class GitHubAuthAPI(GitHubBaseAPI):
    """API client for GitHub authentication endpoints."""

    def __init__(self) -> None:
        """Initialize the GitHub API client."""
        self.base_url = "https://github.com"
        self.session = None
        self.headers = {"Accept": "application/json"}

    async def ensure_session(self) -> None:
        """Create session if needed."""
        if self.session is None or self.session.closed:
            self.session = aiohttp.ClientSession(headers=self.headers)

    async def close(self) -> None:
        """Close session if it exists."""
        if self.session and not self.session.closed:
            await self.session.close()
            self.session = None

    async def start_device_flow(self, client_id: str, scope: str) -> DeviceFlowInfo:
        """Start the device flow authentication process."""
        await self.ensure_session()
        response = await self.make_request(
            "POST",
            f"{self.base_url}/login/device/code",
            json={"client_id": client_id, "scope": scope}
        )

        if not response:
            msg = "Failed to start device flow authentication"
            raise GitHubAuthError(
                msg,
                error_type="device_flow_start_failed"
            )

        return DeviceFlowInfo.model_validate(response)

    async def poll_for_token(self, client_id: str, device_code: str) -> TokenResponse:
        """Poll GitHub for the access token using device code."""
        await self.ensure_session()
        response = await self.make_request(
            "POST",
            f"{self.base_url}/login/oauth/access_token",
            json={
                "client_id": client_id,
                "device_code": device_code,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            }
        )

        if not response:
            msg = "Failed to poll for access token"
            raise GitHubAuthError(
                msg,
                error_type="token_poll_failed"
            )

        if "error" in response:
            if response["error"] == "authorization_pending":
                return TokenResponse(success=False)
            raise GitHubAuthError(
                message=f"Authentication error: {response['error']}",
                error_type=response["error"]
            )

        return TokenResponse(
            success=True,
            access_token=response["access_token"]
        )


class GitHubAuthManager:
    """Handles GitHub authentication flows."""

    def __init__(self, client_id: str) -> None:
        """Initialize the authentication manager."""
        self.client_id = client_id
        self._auth_api = None

    async def authenticate(self, scope: str) -> str:
        """Complete authentication flow and return access token."""
        auth_api = GitHubAuthAPI()

        try:
            flow_info = await auth_api.start_device_flow(self.client_id, scope)
            logger.info(
                "\n**********************************************************************************\n")
            logger.info("Device flow started. Please complete authentication:")
            logger.info("Verification URL: %s", flow_info.verification_uri)
            logger.info("User code: %s", flow_info.user_code)
            logger.info("Code expires in %d minutes",
                        flow_info.expires_in // 60)
            logger.info(
                "\n**********************************************************************************\n")
            start_time = time.time()
            timeout = flow_info.expires_in - 5  # 5 seconds buffer

            while time.time() - start_time < timeout:
                try:
                    token_response = await auth_api.poll_for_token(self.client_id, flow_info.device_code)
                    if token_response.success and token_response.access_token:
                        logger.info("Authentication successful")
                        return token_response.access_token
                except GitHubAuthError as e:
                    if e.error_type != "authorization_pending":
                        logger.exception("Authentication error!")
                        raise
                    # Ignore authorization_pending errors and continue polling

                await asyncio.sleep(flow_info.interval)

            logger.error("Authentication timed out after %d minutes",
                         timeout)
            raise AuthenticationTimeoutError(timeout)  # noqa: TRY301

        except Exception as e:
            if not isinstance(e, GitHubError):
                # Wrap non-GitHub errors
                logger.exception("Unexpected error during authentication")
                msg = f"Authentication failed: {e!s}"
                raise GitHubAuthError(msg) from e
            raise
        finally:
            await auth_api.close()



"""GitHub API service for managing Git content within a repository."""

import base64
import hashlib
import logging

from ohc_backend.services.github.rest import GitHubRestAPI

from .models import (
    CommitFilesRequest,
    GitBlob,
    GitCommit,
    GitReference,
    GitTree,
    GitTreeItem,
)

logger = logging.getLogger(__name__)


class GitHubContentManager:
    """Handles Git content operations within a repository."""

    def __init__(self, api: GitHubRestAPI, repository_full_name: str) -> None:
        """Initialize the content manager."""
        self.api = api
        self.repository_full_name = repository_full_name
        self.base_url = f"{api.base_url}/repos/{repository_full_name}"

    async def create_blob(self, content: str) -> GitBlob:
        """Create a Git blob from content."""
        logger.debug("Creating new blob for file content")
        data = {
            "content": base64.b64encode(content.encode()).decode(),
            "encoding": "base64",
        }
        response = await self.api.make_request(
            "POST",
            f"{self.base_url}/git/blobs",
            json=data,
        )
        blob = GitBlob(**response)
        logger.debug("Created blob with SHA: %s", blob.sha)
        return blob

    async def create_tree(self, items: list[GitTreeItem]) -> GitTree:
        """Create a Git tree containing multiple items."""
        logger.debug("Creating new tree with %d items", len(items))
        data = {"tree": [item.model_dump(exclude_none=True) for item in items]}
        response = await self.api.make_request(
            "POST",
            f"{self.base_url}/git/trees",
            json=data,
        )
        tree = GitTree(**response)
        logger.debug("Created tree with SHA: %s", tree.sha)
        return tree

    async def get_tree(self, tree_sha: str, *, recursive: bool = False) -> GitTree:
        """Get a Git tree by its SHA."""
        logger.debug("Fetching tree with SHA: %s (recursive: %s)",
                     tree_sha, recursive)
        params = {"recursive": "1"} if recursive else {}
        response = await self.api.make_request(
            "GET",
            f"{self.base_url}/git/trees/{tree_sha}",
            params=params,
        )
        tree = GitTree(**response)
        logger.debug("Retrieved tree containing %d items", len(tree.tree))
        return tree

    async def create_commit(
        self,
        tree_sha: str,
        message: str,
        parent_sha: str,
    ) -> GitCommit:
        """Create a Git commit pointing to a tree."""
        logger.debug("Creating new commit for tree: %s", tree_sha)
        data = {"message": message, "tree": tree_sha, "parents": [parent_sha]}
        response = await self.api.make_request(
            "POST",
            f"{self.base_url}/git/commits",
            json=data,
        )
        commit = GitCommit(**response)
        logger.debug("Created commit with SHA: %s", commit.sha)
        return commit

    async def get_branch_reference(self, branch: str = "main") -> GitReference:
        """Get the current commit SHA for a branch."""
        logger.debug("Fetching reference for branch: %s", branch)
        response = await self.api.make_request(
            "GET",
            f"{self.base_url}/git/refs/heads/{branch}",
        )
        ref = GitReference(**response)
        logger.debug("Branch %s is at commit: %s", branch, ref.object["sha"])
        return ref

    async def update_branch_reference(
        self,
        branch: str,
        commit_sha: str,
    ) -> GitReference:
        """Update branch to point to a new commit."""
        logger.debug("Updating branch %s to point to commit: %s",
                     branch, commit_sha)
        url = f"{self.base_url}/git/refs/heads/{branch}"
        data = {"sha": commit_sha, "force": True}
        response = await self.api.make_request("PATCH", url, json=data)
        ref = GitReference(**response)
        logger.debug("Successfully updated branch %s", branch)
        return ref

    async def commit_files(
        self,
        request: CommitFilesRequest,
        ref: GitReference = None,
        tree: GitTree = None
    ) -> GitCommit:
        """Commit multiple files to a branch."""
        logger.debug(
            "Starting commit of %d files to branch: %s",
            len(request.files),
            request.branch,
        )
        try:
            if ref is None or tree is None:
                # Only fetch if not provided
                current_ref = await self.get_branch_reference(request.branch)
                current_tree = await self.get_tree(current_ref.object["sha"])
            else:
                # Use provided ref and tree
                current_ref = ref
                current_tree = tree

            tree_items = []
            if request.update_only:
                logger.debug(
                    "Preserving existing files not included in this commit")
                tree_items.extend(
                    item for item in current_tree.tree if item.path not in request.files)

            logger.debug("Creating blobs for new/updated files")
            for path, content in request.files.items():
                logger.debug("Processing file: %s", path)
                blob = await self.create_blob(content)
                tree_items.append(GitTreeItem(path=path, sha=blob.sha))

            logger.debug("Creating new tree with %d total items",
                         len(tree_items))
            new_tree = await self.create_tree(tree_items)

            logger.debug("Creating commit with the new tree")
            commit = await self.create_commit(
                new_tree.sha,
                request.message,
                current_ref.object["sha"],
            )

            logger.debug("Updating branch %s to the new commit",
                         request.branch)
            await self.update_branch_reference(request.branch, commit.sha)

            logger.debug("Files successfully committed")

        except Exception:
            logger.exception("Failed to commit files: %s")
            raise
        else:
            return commit

    def calculate_blob_sha(self, content: str) -> str:
        """Calculate the SHA-1 hash that GitHub would assign to this content."""
        content_bytes = content.encode("utf-8")
        header = f"blob {len(content_bytes)}\0".encode("ascii")
        full_blob = header + content_bytes
        return hashlib.sha1(full_blob).hexdigest()  # noqa: S324

    async def get_changed_files(self, files: dict[str, str], branch: str = "main") -> dict[str, str]:
        """Determine which files have actually changed compared to the repository."""
        logger.debug("Checking for changes in %d files", len(files))

        # Get current tree
        current_ref = await self.get_branch_reference(branch)
        current_tree = await self.get_tree(current_ref.object["sha"], recursive=True)

        # Create path to SHA mapping of current tree
        current_shas = {item.path: item.sha for item in current_tree.tree}

        # Compare SHAs to find changes
        changed_files = {}
        for path, content in files.items():
            new_sha = self.calculate_blob_sha(content)
            if path not in current_shas or current_shas[path] != new_sha:
                changed_files[path] = content

        logger.debug("Found %d changed files", len(changed_files))
        return changed_files, current_ref, current_tree

    async def commit_changed_files(
        self,
        files: dict[str, str],
        message: str,
        branch: str = "main",
    ) -> GitCommit | None:
        """Commit only files that have actually changed."""
        result = await self.get_changed_files(files, branch)
        changed_files, current_ref, current_tree = result
        logger.debug("Changed files: %s", list(changed_files.keys()))

        if not changed_files:
            logger.debug("No files have changed, skipping commit")
            return None

        request = CommitFilesRequest(
            files=changed_files,
            message=message,
            branch=branch,
            update_only=True,
        )

        return await self.commit_files(request, current_ref, current_tree)

    async def get_file_contents(self, path: str) -> str | None:
        """Fetch a file from the repository if it exists."""
        logger.debug("Checking for file %s in repository: %s",
                     path, self.repository_full_name)

        response = await self.api.make_request(
            "GET",
            f"{self.base_url}/contents/{path}",
        )
        if response is None:  # 404 Not Found
            logger.debug("File %s not found", path)
            return None

        return base64.b64decode(response["content"]).decode("utf-8")



"""Tests for OHCState class."""

from datetime import datetime

import pytest

from ohc_backend.models.ha_entity import Automation, Script
from ohc_backend.services.ohc_state import OHCState


@pytest.fixture
def ohc_state() -> OHCState:
    """Fixture for creating OHCState."""
    return OHCState()


@pytest.fixture
def automation() -> Automation:
    """Fixture for creating an Automation entity."""
    last_changed_iso = "2025-02-08T19:27:24.580521+00:00"
    return Automation(
        automation_id="234234234234234",
        entity_id="automation.test",
        friendly_name="Test Automation",
        state="on",
        last_changed=datetime.fromisoformat(last_changed_iso),
    )


@pytest.fixture
def script() -> Script:
    """Fixture for creating a Script entity."""
    last_changed_iso = "2025-02-08T19:27:24.580521+00:00"
    return Script(
        entity_id="script.test",
        friendly_name="Test Script",
        state="on",
        last_changed=datetime.fromisoformat(last_changed_iso),
    )


def test_create_entity(ohc_state: OHCState, automation: Automation) -> None:
    """Test creating an entity."""
    entity = ohc_state.add(automation)
    assert entity == automation
    assert ohc_state.get_entity(automation.entity_id) == automation


def test_get_entity(ohc_state: OHCState, automation: Automation) -> None:
    """Test getting an entity."""
    ohc_state.add(automation)
    entity = ohc_state.get_entity(automation.entity_id)
    assert entity == automation


def test_get_automation(ohc_state: OHCState, automation: Automation) -> None:
    """Test getting an automation entity."""
    ohc_state.add(automation)
    entity = ohc_state.get_automation(automation.entity_id)
    assert entity == automation


def test_get_script(ohc_state: OHCState, script: Script) -> None:
    """Test getting a script entity."""
    ohc_state.add(script)
    entity = ohc_state.get_script(script.entity_id)
    assert entity == script


def test_get_automations(ohc_state: OHCState, automation: Automation) -> None:
    """Test getting all automation entities."""
    ohc_state.add(automation)
    automations = ohc_state.get_automations()
    assert automations == [automation]


def test_get_scripts(ohc_state: OHCState, script: Script) -> None:
    """Test getting all script entities."""
    ohc_state.add(script)
    scripts = ohc_state.get_scripts()
    assert scripts == [script]


def test_get_entities(ohc_state: OHCState, automation: Automation, script: Script) -> None:
    """Test getting all entities."""
    ohc_state.add(automation)
    ohc_state.add(script)
    entities = ohc_state.get_entities()
    assert entities == [automation, script]


def test_update_entity(ohc_state: OHCState, automation: Automation) -> None:
    """Test updating an entity."""
    ohc_state.add(automation)
    updated_automation = Automation(
        automation_id="234234234234234",
        entity_id="automation.test",
        friendly_name="Updated Automation",
        state=automation.state,
        last_changed=automation.last_changed,
    )
    entity = ohc_state.update(updated_automation)
    assert entity == updated_automation
    assert ohc_state.get_entity(automation.entity_id) == updated_automation


def test_delete_entity(ohc_state: OHCState, automation: Automation) -> None:
    """Test deleting an entity."""
    ohc_state.add(automation)
    result = ohc_state.remove(automation.entity_id)
    assert result is True
    assert ohc_state.get_entity(automation.entity_id) is None


def test_upsert_entity(ohc_state: OHCState, automation: Automation) -> None:
    """Test upserting an entity."""
    entity = ohc_state.upsert(automation)
    assert entity == automation
    assert ohc_state.get_entity(automation.entity_id) == automation

    updated_automation = Automation(
        automation_id="234234234234234",
        entity_id="automation.test",
        friendly_name="Updated Automation",
        state=automation.state,
        last_changed=automation.last_changed,
    )
    entity = ohc_state.upsert(updated_automation)
    assert entity == updated_automation
    assert ohc_state.get_entity(automation.entity_id) == updated_automation


def test_to_json(ohc_state: OHCState, automation: Automation, script: Script) -> None:
    """Test converting state to JSON."""
    ohc_state.add(automation)
    ohc_state.add(script)
    json_str = ohc_state.to_json()
    assert isinstance(json_str, str)


def test_from_json(ohc_state: OHCState, automation: Automation, script: Script) -> None:
    """Test creating state from JSON."""
    ohc_state.add(automation)
    ohc_state.add(script)
    json_str = ohc_state.to_json()

    new_ohc_state = OHCState.from_json(json_str)
    assert new_ohc_state.get_entity(automation.entity_id) == automation
    assert new_ohc_state.get_entity(script.entity_id) == script



"""Main FastAPI application file."""

import logging
import os
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

from fastapi import FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from ohc_backend.errors import AppError, ErrorCode
from ohc_backend.services.github.errors import GitHubAuthError
from ohc_backend.utils.logging import configure_logging, log_error
from ohc_backend.utils.request_context import request_id_middleware

from .dependencies import deps
from .routers import automations

configure_logging()

logger = logging.getLogger(__name__)


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator:  # noqa: ARG001
    """Application lifespan context manager."""
    try:
        logger.info("Initializing application dependencies")
        await deps.async_init()

        try:
            logger.info("Setting up GitHub integration")
            await deps.setup_github()
        except TimeoutError as e:
            msg = "GitHub authentication timed out. Please restart the add-on to try again."
            log_error(logger, msg, e, critical=True)
            raise AppError(message=msg, status_code=status.HTTP_504_GATEWAY_TIMEOUT,
                           error_code=ErrorCode.AUTHENTICATION_FAILED) from e
        except GitHubAuthError as e:
            msg = "Failed to authenticate with GitHub"
            log_error(logger, msg, e, critical=True)
            raise AppError(message=msg, status_code=status.HTTP_401_UNAUTHORIZED,
                           error_code=ErrorCode.AUTHENTICATION_FAILED) from e
        except Exception as e:
            msg = f"Failed to set up GitHub integration: {e}"
            log_error(logger, msg, e, critical=True)
            raise AppError(message=msg, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                           error_code=ErrorCode.GITHUB_API_ERROR) from e

        try:
            logger.info("Starting sync manager")
            sync_manager = deps.get_sync_manager()
            await sync_manager.start()
            logger.info("Application started successfully")
            yield
        except Exception as e:
            log_error(logger, "Error starting sync manager", e, critical=True)
            raise AppError(
                message=f"Sync manager failed to start: {e!s}", error_code=ErrorCode.SYNC_ERROR) from e
    except Exception as e:
        if not isinstance(e, AppError):
            log_error(logger, "Critical startup error", e, critical=True)
            msg = "Application failed to start!"
            raise AppError(msg) from e
        raise
    finally:
        logger.info("Cleaning up application resources")
        await deps.cleanup()


app = FastAPI(lifespan=lifespan, debug=os.getenv("ENVIRONMENT") == "dev")
app.middleware("http")(request_id_middleware)


@app.exception_handler(AppError)
async def app_error_handler(request: Request, exc: AppError) -> JSONResponse:  # noqa: ARG001
    """Handle application errors."""
    # No need to log here, already logged where the error occurred
    return JSONResponse(
        status_code=exc.status_code,
        content=exc.to_dict()
    )


@app.exception_handler(Exception)
async def general_exception_handler(request: Request, exc: Exception) -> JSONResponse:
    """Handle unhandled exceptions."""
    log_error(
        logger, f"Unhandled exception in {request.method} {request.url.path}", exc, critical=True)

    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            "error_code": ErrorCode.INTERNAL_ERROR.value,
            "message": "An unexpected error occurred",
            "details": {"type": type(exc).__name__}
        }
    )

app.add_middleware(
    CORSMiddleware,
    allow_origins=["http://localhost:5173"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.include_router(
    automations.router,
    prefix="/api/automations",
    tags=["automations"],
)
# app.include_router(
#     github.router,
#     prefix="/api/github",
#     tags=["github"],
# )


@app.get("/api/health")
async def health_check() -> dict:
    """Endpoint to check the health status of the application."""
    return {"status": "ok"}


# Mount frontend last (catches all other routes)
# Only serve frontend files in production
# if os.getenv("ENVIRONMENT") != "dev":
#     app.mount("/", StaticFiles(directory="../frontend/build",
#               html=True), name="frontend")



"""Dependencies for FastAPI."""

import logging
import os

from ohc_backend.services.github import GitHubClient
from ohc_backend.services.ha_service import HomeAssistantService
from ohc_backend.services.ohc_state import OHCState
from ohc_backend.services.settings import Settings
from ohc_backend.services.sync_manager import SyncManager

logger = logging.getLogger(__name__)


class DependencyManager:
    """Dependency Manager."""

    def __init__(self, settings_path: str) -> None:
        """Initialize dependency manager."""
        self._settings_path: str = settings_path
        self._settings: Settings | None = None
        self._github_client: GitHubClient | None = None
        self._sync_manager: SyncManager | None = None
        self._home_assistant_service: HomeAssistantService | None = None

    async def async_init(self) -> None:
        """Async initializer."""
        logger.info("Initializing setting from path: %s", self._settings_path)
        self._settings = Settings(self._settings_path)
        await self._settings.async_init()

    async def cleanup(self) -> None:
        """Cleanup dependency resources."""
        if self._sync_manager:
            await self._sync_manager.stop()

        if self._github_client:
            await self._github_client.close()

        if self._home_assistant_service:
            await self._home_assistant_service.close()

    def get_github_client(self) -> GitHubClient:
        """Get Github client."""
        if not self._github_client:
            api_url = self.get_settings().gh_config.api_url
            client_id = self.get_settings().gh_config.client_id
            token = self.get_settings().gh_token
            self._github_client = GitHubClient(
                api_url=api_url, client_id=client_id, access_token=token)
        return self._github_client

    def get_settings(self) -> Settings:
        """Get settings manager."""
        return self._settings

    def get_ha_service(self) -> HomeAssistantService:
        """Get Home Assistant Service."""
        if not self._home_assistant_service:
            settings = self.get_settings()
            server = settings.ha_config.server
            token = settings.ha_config.token
            self._home_assistant_service = HomeAssistantService(server, token)
        return self._home_assistant_service

    def get_sync_manager(self) -> SyncManager:
        """Get sync manager."""
        if not self._sync_manager:
            config = self.get_settings().sync_config
            self._sync_manager = SyncManager(
                self.get_ha_service(), self.get_github_client(), config)
        return self._sync_manager

    def get_ohc_state(self) -> OHCState:
        """Get state manager."""
        return self.get_sync_manager().get_ohc_state()

    async def clear_github_token(self) -> None:
        """Clear GitHub token in settings when authentication fails."""
        if self._settings and self._settings.gh_token:
            logger.warning(
                "GitHub authentication failed. Clearing invalid token.")
            self._settings.gh_config.access_token = None
            await self._settings.save()

            # Also update the token in the existing client if it exists
            if self._github_client:
                self._github_client.rest_api.set_auth_token("")

    async def setup_github(self) -> None:
        """Start the github auth flow and create repo if required."""
        settings = self.get_settings()
        client = self.get_github_client()
        if not settings.gh_token:
            scope = settings.gh_config.scope
            token = await client.authenticate(scope)
            settings.gh_config.access_token = token
            await settings.save()
            client.rest_api.set_auth_token(token)

        await client.init_repository(settings.gh_config.repo_request)


data_folder = os.getenv("HA_DATA_FOLDER", "/data")
deps = DependencyManager(f"{data_folder}/config.json")



"""Github API router."""


# router = APIRouter()
# logger = logging.getLogger(__name__)

# setup_status = {}


# @router.post("/device-code")
# async def start_github_auth(
#     settings: Annotated[Settings, Depends(deps.get_settings)],
#     github_auth_service: Annotated[GithubAuthService, Depends(deps.get_github_auth_service)],
# ) -> DeviceFlowInfo:
#     """Start the GitHub device flow authentication process."""
#     return await github_auth_service.start_device_flow(settings.gh_auth_config.scope)


# @router.get("/poll-token/{device_code}")
# async def check_auth_status(
#     device_code: str,
#     settings: Annotated[Settings, Depends(deps.get_settings)],
#     github_auth_service: Annotated[GithubAuthService, Depends(deps.get_github_auth_service)],
# ) -> dict:
#     """Check the status of GitHub authentication."""
#     token_response = await github_auth_service.poll_for_token(device_code)
#     # TODO @<danieldotnl>: remove this logging  # noqa: FIX002, TD003
#     logger.debug("Token response: %s", token_response)

#     if token_response.success:
#         settings.gh_config.access_token = token_response.access_token
#         await settings.save()

#         sync_manager = deps.get_sync_manager()
#         await sync_manager.start()
#     return {"success": token_response.success}



"""Request context utilities."""
import contextvars
import uuid
from collections.abc import Callable

from fastapi import Request, Response

# Context variable to store request ID
request_id_var = contextvars.ContextVar("request_id", default=None)


def get_request_id() -> str:
    """Get the current request ID."""
    return request_id_var.get()


async def request_id_middleware(request: Request, call_next: Callable[[Request], Response]) -> Response:
    """Middleware to assign a unique ID to each request."""
    request_id = str(uuid.uuid4())[:8]  # Use first 8 chars for brevity
    token = request_id_var.set(request_id)
    try:
        response = await call_next(request)
        response.headers["X-Request-ID"] = request_id
        return response
    finally:
        request_id_var.reset(token)



import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from ohc_backend.models.ha_entity import Automation, HAEntity, Script
from ohc_backend.services.github import GitHubClient
from ohc_backend.services.github.errors import GitHubAPIError, GitHubAuthError, GitHubNotFoundError
from ohc_backend.services.ha_service import HomeAssistantError, HomeAssistantService
from ohc_backend.services.ohc_state import OHCState
from ohc_backend.services.settings import SyncManagerConfig
from ohc_backend.services.sync_manager import SyncManager, SyncStatus


@pytest.fixture
def ha_service():
    """Create a mock HomeAssistantService."""
    return AsyncMock(spec=HomeAssistantService)


@pytest.fixture
def github_client():
    """Create a mock GitHubClient with content manager."""
    client = AsyncMock(spec=GitHubClient)
    client.content = AsyncMock()
    return client


@pytest.fixture
def sync_config():
    """Create a SyncManagerConfig with test settings."""
    return SyncManagerConfig(
        interval=5,  # Short interval for tests
        state_file=".ohcstate/test_state.json",
        ha_max_parallel_requests=5
    )


@pytest.fixture
def sync_manager(ha_service, github_client, sync_config):
    """Create a SyncManager with mock dependencies."""
    manager = SyncManager(ha_service, github_client, sync_config)
    # Ensure the manager is using our mocked client
    assert manager.github is github_client
    return manager


@pytest.fixture
def example_automation():
    """Create an example automation entity."""
    return Automation(
        entity_id="automation.test_automation",
        automation_id="1234",
        friendly_name="Test Automation",
        state="on",
        last_changed="2023-01-01T00:00:00.000000+00:00"
    )


@pytest.fixture
def example_script():
    """Create an example script entity."""
    return Script(
        entity_id="script.test_script",
        friendly_name="Test Script",
        state="off",
        last_changed="2023-01-01T00:00:00.000000+00:00"
    )


@pytest.mark.asyncio
async def test_start_loads_state_from_github(sync_manager, github_client):
    """Test that starting the sync manager loads state from GitHub."""
    # Setup mock for GitHub content
    test_state = OHCState()
    test_automation = Automation(
        entity_id="automation.test_auto",
        automation_id="123",
        friendly_name="Test Auto",
        state="on",
        last_changed="2023-01-01T00:00:00.000000+00:00"
    )
    test_state.add(test_automation)
    state_json = test_state.to_json()

    github_client.content.get_file_contents.return_value = state_json

    # Start the sync manager
    await sync_manager.start()

    # Verify GitHub was called to get state file
    github_client.content.get_file_contents.assert_called_once_with(
        sync_manager.sync_config.state_file
    )

    # Verify state was loaded
    assert len(sync_manager.get_ohc_state().get_entities()) == 1
    assert sync_manager.get_ohc_state().get_entity(
        "automation.test_auto") is not None


@pytest.mark.asyncio
async def test_start_handles_not_found_error(sync_manager, github_client):
    """Test that the sync manager handles state file not found on GitHub."""
    # Setup mock for GitHub content to raise not found error
    github_client.content.get_file_contents.side_effect = GitHubNotFoundError(
        "file", sync_manager.sync_config.state_file)

    # Start the sync manager
    await sync_manager.start()

    # Verify GitHub was called to get state file
    github_client.content.get_file_contents.assert_called_once_with(
        sync_manager.sync_config.state_file
    )

    # Verify empty state was created
    assert len(sync_manager.get_ohc_state().get_entities()) == 0
    assert sync_manager.status == SyncStatus.RUNNING


@pytest.mark.asyncio
async def test_start_handles_auth_error(sync_manager, github_client):
    """Test that the sync manager handles GitHub auth errors."""
    # Setup mock for GitHub content to raise auth error
    github_client.content.get_file_contents.side_effect = GitHubAuthError(
        "Auth failed")

    # Start should raise RuntimeError when auth fails
    with pytest.raises(RuntimeError):
        await sync_manager.start()

    # Verify status is set to ERROR
    assert sync_manager.status == SyncStatus.ERROR


@pytest.mark.asyncio
async def test_stop_cancels_tasks(sync_manager, github_client):
    """Test that stopping the sync manager cancels all tasks."""
    # Configure the mock to return None to simulate file not found
    github_client.content.get_file_contents.return_value = None

    # First start the sync manager
    with patch.object(sync_manager, "_async_loop", return_value=asyncio.Future()):
        await sync_manager.start()

        # Manually create a sleep task for testing
        sync_manager._sleep_task = asyncio.create_task(asyncio.sleep(100))

        # Now stop it
        await sync_manager.stop()

        # Verify sleep task was cancelled
        assert sync_manager._sleep_task.cancelled()
        assert sync_manager.status == SyncStatus.STOPPED


@pytest.mark.asyncio
async def test_fetch_entity_contents_parallel(sync_manager, example_automation, example_script):
    """Test fetching entity contents in parallel with correct concurrency."""
    # Setup _fetch_entity_content to return different content for different entities
    async def mock_fetch_content(entity):
        if entity.entity_id == example_automation.entity_id:
            return "automation: content"
        return "script: content"

    sync_manager._fetch_entity_content = AsyncMock(
        side_effect=mock_fetch_content)

    # Test with multiple entities
    entities = [example_automation, example_script]
    results = await sync_manager.fetch_entity_contents_parallel(entities)

    # Verify results
    assert len(results) == 2
    assert results[0][0] == example_automation
    assert results[0][1] == "automation: content"
    assert results[1][0] == example_script
    assert results[1][1] == "script: content"

    # Verify _fetch_entity_content was called for each entity
    assert sync_manager._fetch_entity_content.call_count == 2


@pytest.mark.asyncio
async def test_fetch_entity_contents_handles_errors(sync_manager, example_automation, example_script):
    """Test fetching entity contents handles errors properly."""
    # Setup _fetch_entity_content to raise an error for one entity
    async def mock_fetch_content(entity):
        if entity.entity_id == example_automation.entity_id:
            raise HomeAssistantError("Failed to fetch")
        return "script: content"

    sync_manager._fetch_entity_content = AsyncMock(
        side_effect=mock_fetch_content)

    # Test with multiple entities should raise RuntimeError
    entities = [example_automation, example_script]
    with pytest.raises(RuntimeError):
        await sync_manager.fetch_entity_contents_parallel(entities)


@pytest.mark.asyncio
async def test_fetch_entity_content(sync_manager, ha_service, example_automation, example_script):
    """Test fetching content for different entity types."""
    # Setup mocks for ha_service
    ha_service.get_automation_content.return_value = "automation yaml"
    ha_service.get_script_content.return_value = "script yaml"

    # Test automation
    content = await sync_manager._fetch_entity_content(example_automation)
    assert content == "automation yaml"
    ha_service.get_automation_content.assert_called_once_with(
        example_automation.automation_id)

    # Test script
    content = await sync_manager._fetch_entity_content(example_script)
    assert content == "script yaml"
    ha_service.get_script_content.assert_called_once_with(
        example_script.entity_id)


@pytest.mark.asyncio
async def test_identify_potential_changes(sync_manager, example_automation):
    """Test identification of potential changes in entities."""
    # Create a state with an existing entity
    state_copy = OHCState()
    existing_automation = Automation(
        entity_id="automation.existing",
        automation_id="existing",
        friendly_name="Existing Automation",
        state="on",
        last_changed="2023-01-01T00:00:00.000000+00:00"
    )
    state_copy.add(existing_automation)

    # Create an updated version of the existing entity
    updated_automation = Automation(
        entity_id="automation.existing",
        automation_id="existing",
        friendly_name="Updated Name",  # Name changed
        state="on",
        last_changed="2023-01-02T00:00:00.000000+00:00"  # Time changed
    )

    # Also add a new entity and an entity to be marked as deleted
    deleted_automation = Automation(
        entity_id="automation.to_delete",
        automation_id="to_delete",
        friendly_name="To Delete",
        state="on",
        last_changed="2023-01-01T00:00:00.000000+00:00"
    )
    state_copy.add(deleted_automation)

    # Entities from HA - includes updated and example but not deleted
    ha_entities = [updated_automation, example_automation]

    # Run the identification process
    result = await sync_manager._identify_potential_changes(ha_entities, state_copy)

    # Unpack results
    changed, inserted, deleted = result

    # Verify results
    assert len(changed) == 1
    assert changed[0].entity_id == "automation.existing"
    assert changed[0].friendly_name == "Updated Name"

    assert len(inserted) == 1
    assert inserted[0].entity_id == example_automation.entity_id

    assert len(deleted) == 1
    assert deleted[0].entity_id == "automation.to_delete"
    assert deleted[0].is_deleted  # Should be marked as deleted


@pytest.mark.asyncio
async def test_check_content_changes(sync_manager, github_client, example_automation):
    """Test checking which entities have actual content changes."""
    # Make sure the sync_manager uses our mocked client
    sync_manager.github = github_client

    # Setup state copy
    state_copy = OHCState()

    # Reset any previous mock configurations
    github_client.content.get_file_contents.reset_mock()

    # Setup GitHub client content methods
    github_client.content.get_file_contents.return_value = "old content"

    # Mock the get_changed_files method which is causing the exception
    changed_files_result = (
        {"automations/automation.test_automation.yaml": "new content"}, None, None)
    github_client.content.get_changed_files = AsyncMock(
        return_value=changed_files_result)

    # Create a parallel fetch result
    content_results = [(example_automation, "new content")]
    sync_manager.fetch_entity_contents_parallel = AsyncMock(
        return_value=content_results)

    # Test with a changed entity
    changed_entities = [example_automation]
    inserted_entities = []

    # Run the method being tested
    result = await sync_manager._check_content_changes(changed_entities, inserted_entities, state_copy)

    # Verify results
    assert result is not None
    files, final_changed = result

    # Verify the correct file was identified as changed
    file_path = f"automations/{example_automation.entity_id}.yaml"
    assert file_path in files
    assert files[file_path] == "new content"
    assert len(final_changed) == 1
    assert final_changed[0] == example_automation


@pytest.mark.asyncio
async def test_process_changes_integration(sync_manager, github_client, example_automation):
    """Test the full process_changes flow integrating all sub-processes."""
    # Setup mock for identifying potential changes
    ha_entities = [example_automation]
    state_copy = OHCState()

    # Setup mocks
    with patch.object(sync_manager, "_identify_potential_changes") as mock_identify:
        # Return a mix of changed, inserted, and deleted entities
        mock_identify.return_value = ([example_automation], [], [])

        # Setup content check mock
        with patch.object(sync_manager, "_check_content_changes") as mock_content:
            # Return files and final changed entities
            test_files = {
                "automations/automation.test_automation.yaml": "content"}
            mock_content.return_value = (test_files, [example_automation])

            # Run process_changes
            result = await sync_manager._process_changes(ha_entities, state_copy)

            # Verify results
            files, updated, inserted, deleted = result
            assert sync_manager.sync_config.state_file in files
            assert "automations/automation.test_automation.yaml" in files
            assert len(updated) == 1
            assert len(inserted) == 0
            assert len(deleted) == 0


@pytest.mark.asyncio
async def test_commit_changes_success(sync_manager, github_client):
    """Test successfully committing changes to GitHub."""
    # Setup mock for commit_changed_files
    github_client.content.commit_changed_files.return_value = True

    # Test files
    files = {
        "automations/test.yaml": "content",
        sync_manager.sync_config.state_file: "{}"
    }

    # Commit the changes
    result = await sync_manager._commit_changes(files, 1, 1, 0)

    # Verify result
    assert result is True
    github_client.content.commit_changed_files.assert_called_once()


@pytest.mark.asyncio
async def test_commit_changes_auth_error(sync_manager, github_client):
    """Test handling auth errors when committing changes."""
    # Setup mock to raise auth error
    github_client.content.commit_changed_files.side_effect = GitHubAuthError(
        "Auth failed")

    # Test files
    files = {"automations/test.yaml": "content"}

    # Commit the changes
    result = await sync_manager._commit_changes(files, 1, 0, 0)

    # Verify result
    assert result is False
    assert sync_manager.status == SyncStatus.ERROR


@pytest.mark.asyncio
async def test_commit_changes_api_error(sync_manager, github_client):
    """Test handling API errors when committing changes."""
    # Set the initial status to RUNNING
    sync_manager.status = SyncStatus.RUNNING

    # Make sure sync_manager is using our mocked client
    sync_manager.github = github_client

    # Setup mock to raise API error
    github_client.content.commit_changed_files.side_effect = GitHubAPIError(
        "API Error", status_code=500)

    # Test files
    files = {"automations/test.yaml": "content"}

    # Commit the changes
    result = await sync_manager._commit_changes(files, 1, 0, 0)

    # Verify result
    assert result is False
    # Status should not change to ERROR for temporary API issues
    assert sync_manager.status == SyncStatus.RUNNING


@pytest.mark.asyncio
async def test_run_integration(sync_manager):
    """Test the full sync run integration."""
    # Setup mocks for all key methods in the run flow
    with patch.object(sync_manager, "_fetch_entities_and_prepare_state") as mock_fetch:
        ha_entities = [MagicMock(spec=HAEntity)]
        state_copy = OHCState()
        mock_fetch.return_value = (ha_entities, state_copy)

        with patch.object(sync_manager, "_process_changes") as mock_process:
            files = {"test.yaml": "content"}
            updated = [MagicMock(spec=HAEntity)]
            inserted = [MagicMock(spec=HAEntity)]
            deleted = []
            mock_process.return_value = (files, updated, inserted, deleted)

            with patch.object(sync_manager, "_commit_changes") as mock_commit:
                mock_commit.return_value = True

                # Run the sync
                await sync_manager.run()

                # Verify all methods were called with expected arguments
                mock_fetch.assert_called_once()
                mock_process.assert_called_once_with(ha_entities, state_copy)
                mock_commit.assert_called_once_with(
                    files, len(updated), len(inserted), len(deleted))

                # State should be updated on successful commit
                assert sync_manager._ohc_state == state_copy


@pytest.mark.asyncio
async def test_run_no_changes(sync_manager):
    """Test sync run when no changes are detected."""
    # Setup mocks for all key methods in the run flow
    with patch.object(sync_manager, "_fetch_entities_and_prepare_state") as mock_fetch:
        ha_entities = [MagicMock(spec=HAEntity)]
        state_copy = OHCState()
        mock_fetch.return_value = (ha_entities, state_copy)

        with patch.object(sync_manager, "_process_changes") as mock_process:
            files = {}  # Empty files dict
            updated = []
            inserted = []
            deleted = []
            mock_process.return_value = (files, updated, inserted, deleted)

            with patch.object(sync_manager, "_commit_changes") as mock_commit:
                # Run the sync
                await sync_manager.run()

                # Process should be called, but commit should not
                mock_process.assert_called_once()
                mock_commit.assert_not_called()

                # State should still be updated even when no files change
                assert sync_manager._ohc_state == state_copy


@pytest.mark.asyncio
async def test_run_fetch_error(sync_manager):
    """Test sync run handling errors during entity fetching."""
    # Setup mock to return None (indicating error)
    with patch.object(sync_manager, "_fetch_entities_and_prepare_state") as mock_fetch:
        mock_fetch.return_value = None

        with patch.object(sync_manager, "_process_changes") as mock_process:
            # Run the sync
            await sync_manager.run()

            # Process should not be called when fetch fails
            mock_process.assert_not_called()


@pytest.mark.asyncio
async def test_async_loop_handles_cancellation(sync_manager):
    """Test that async loop handles task cancellation properly."""
    # Mock the run method
    sync_manager.run = AsyncMock()

    # Set status to running so the loop will execute
    sync_manager.status = SyncStatus.RUNNING

    # Create a controlled way to exit the loop
    async def mock_sleep(seconds):
        # First call will raise CancelledError to simulate cancellation
        sync_manager.status = SyncStatus.STOPPED
        raise asyncio.CancelledError()

    # Patch the sleep function to cause cancellation
    with patch("asyncio.sleep", mock_sleep):
        # Run the loop - it should exit when mock_sleep raises CancelledError
        await sync_manager._async_loop()

        # Verify run was called once
        sync_manager.run.assert_called_once()


@pytest.mark.asyncio
async def test_async_loop_handles_exceptions(sync_manager):
    """Test that async loop handles exceptions during run()."""
    # Track run count
    run_count = 0

    # Override the run method to raise an exception but also exit the loop
    async def mock_run():
        nonlocal run_count
        run_count += 1
        # After first call, set status to exit the loop
        if run_count >= 1:
            sync_manager.status = SyncStatus.STOPPED
        # Always raise an exception
        raise Exception("Test error")

    # Replace the run method
    sync_manager.run = mock_run

    # Set initial status to running
    sync_manager.status = SyncStatus.RUNNING

    # Run the loop
    await sync_manager._async_loop()

    # Verify the run method was called at least once
    assert run_count >= 1
    # Verify status is now stopped
    assert sync_manager.status == SyncStatus.STOPPED



"""Application error classes."""
from enum import Enum
from typing import Any

from fastapi import status


class ErrorCode(Enum):
    """Application error codes."""

    AUTHENTICATION_FAILED = "auth_failed"
    NOT_FOUND = "not_found"
    VALIDATION_ERROR = "validation_error"
    GITHUB_API_ERROR = "github_api_error"
    HOME_ASSISTANT_ERROR = "ha_error"
    SYNC_ERROR = "sync_error"
    INTERNAL_ERROR = "internal_error"


class AppError(Exception):
    """Base exception for all application errors."""

    def __init__(
        self,
        message: str,
        status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR,
        error_code: ErrorCode = ErrorCode.INTERNAL_ERROR,
        details: dict[str, Any] | None = None
    ) -> None:
        """Initialize the error."""
        self.message = message
        self.status_code = status_code
        self.error_code = error_code
        self.details = details or {}
        super().__init__(message)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict for API responses."""
        return {
            "error_code": self.error_code.value,
            "message": self.message,
            "details": self.details
        }


class AuthenticationError(AppError):
    """Authentication errors."""

    def __init__(self, message: str, details: dict[str, Any] | None = None) -> None:
        """Initialize the error."""
        super().__init__(
            message,
            status_code=status.HTTP_401_UNAUTHORIZED,
            error_code=ErrorCode.AUTHENTICATION_FAILED,
            details=details
        )



"""GitHub-related errors."""


class GitHubError(Exception):
    """Base exception for all GitHub-related errors."""

    def __init__(self, message: str, status_code: int | None = None, error_type: str | None = None) -> None:
        """Initialize the GitHub error."""
        super().__init__(message)
        self.status_code = status_code
        self.error_type = error_type
        self.message = message


class GitHubAuthError(GitHubError):
    """Authentication-related errors."""


class GitHubAPIError(GitHubError):
    """API request errors."""

    def __init__(self, message: str, status_code: int, response_data: dict | None = None, error_type: str = "api_error") -> None:
        """Initialize the API error."""
        super().__init__(message, status_code=status_code, error_type=error_type)
        self.response_data = response_data or {}


class GitHubRateLimitError(GitHubAPIError):
    """Rate limit exceeded errors."""

    def __init__(self, reset_time: int | None = None) -> None:
        """Initialize the rate limit error."""
        super().__init__(
            "GitHub API rate limit exceeded",
            status_code=403,
            error_type="rate_limit"
        )
        self.reset_time = reset_time


class GitHubNotFoundError(GitHubAPIError):
    """Resource not found errors."""

    def __init__(self, resource_type: str, resource_id: str) -> None:
        """Initialize the not found error."""
        super().__init__(
            f"{resource_type} not found: {resource_id}",
            status_code=404,
            error_type="not_found"
        )
        self.resource_type = resource_type
        self.resource_id = resource_id


class GitHubValidationError(GitHubAPIError):
    """Validation errors."""


class GitHubRepositoryError(GitHubError):
    """Repository-specific errors."""


class GitHubContentError(GitHubError):
    """Content operations errors."""



"""Automations API."""

import logging
from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException

from ohc_backend.dependencies import deps
from ohc_backend.models.ha_entity import Automation
from ohc_backend.services.sync_manager import SyncManager

router = APIRouter()
logger = logging.getLogger(__name__)


@router.get("/")
def get_automations(
    sync_manager: Annotated[SyncManager, Depends(deps.get_sync_manager)],
) -> list[Automation]:
    """Get all automations."""
    automations = sync_manager.get_ohc_state().get_automations()
    logger.info("Retrieved %s automations from storage", len(automations))
    return automations


@router.get("/{automation_id}")
def get_automation(
    automation_id: str,
    sync_manager: Annotated[SyncManager, Depends(deps.get_sync_manager)],
) -> Automation:
    """Get a single automation."""
    result = sync_manager.get_ohc_state().get_automation(automation_id)
    if not result:
        raise HTTPException(status_code=404, detail="Automation not found")
    return result



"""GitHub repository management functionality."""

import logging
from pathlib import Path

import aiofiles

from ohc_backend.utils.logging import log_error

from .base import GitHubBaseAPI
from .errors import GitHubNotFoundError
from .models import CommitFilesRequest, GithubRepositoryRequestConfig, Repository

logger = logging.getLogger(__name__)


class GitHubRepositoryManager:
    """Handles repository creation and management."""

    def __init__(self, api: GitHubBaseAPI) -> None:
        """Initialize the repository manager."""
        self.api = api

    async def get_repository(self, full_name: str) -> Repository | None:
        """Fetch repository if it exists."""
        try:
            response = await self.api.make_request(
                "GET",
                f"{self.api.base_url}/repos/{full_name}",
            )
            return Repository(**response)
        except GitHubNotFoundError:
            return None

    async def create_repository(self, config: GithubRepositoryRequestConfig) -> Repository:
        """Create a new repository."""
        data = {
            "name": config.name,
            "private": config.private,
            "auto_init": True,  # GitHub will initialize with a default README
            "description": config.description,
        }

        response = await self.api.make_request(
            "POST",
            f"{self.api.base_url}/user/repos",
            json=data,
        )
        repo = Repository(**response)
        logger.debug("Created repository: %s", repo.full_name)
        return repo

    async def find_repository(self, name: str) -> Repository | None:
        """Find a repository by name for the authenticated user, using GitHub's search API."""
        try:
            # Use the search API to narrow down candidates
            response = await self.api.make_request(
                "GET",
                f"{self.api.base_url}/search/repositories",
                params={
                    # @me refers to authenticated user
                    "q": f"{name} in:name user:@me",
                    "per_page": 100,  # Get enough results to likely include exact match
                },
            )

            logger.debug("Search result count: %s",
                         response.get("total_count"))

            # Filter for exact match on name
            if response.get("items"):
                for repo in response["items"]:
                    if repo.get("name") == name:
                        logger.debug("Found repository: %s",
                                     repo.get("full_name"))
                        return Repository(**repo)
        except Exception as e:
            log_error(logger, "Error finding repository", e)
            return None
        else:
            return None

    async def find_or_create_repository(self, config: GithubRepositoryRequestConfig) -> Repository:
        """Get or create repository."""
        repo = await self.find_repository(config.name)
        if not repo:
            repo = await self.create_repository(config)
        return repo


class RepositoryInitializer:
    """Handles repository initialization with required files."""

    def __init__(self, api: GitHubBaseAPI) -> None:
        """Initialize the repository initializer."""
        self.api = api

    async def initialize_with_readme(self, repo_full_name: str) -> None:
        """Replace the default README with our custom one."""
        from .content import GitHubContentManager

        # Create content manager for this repository
        content_manager = GitHubContentManager(self.api, repo_full_name)

        # Load README template
        template_dir = Path(__file__).parent / "templates"
        readme_path = template_dir / "README.md"

        if readme_path.exists():
            try:
                async with aiofiles.open(readme_path) as f:
                    readme_content = await f.read()

                # Simply update the README - repository already has a commit and main branch
                await content_manager.commit_files(
                    CommitFilesRequest(
                        files={"README.md": readme_content},
                        message="Add README",
                        branch="main",
                        update_only=True  # This will update the existing file
                    )
                )
                logger.info(
                    "Successfully updated repository with OHC README")
            except Exception as e:
                log_error(
                    logger, "Failed to update repository with README", e)
                raise



"""Home Assistant Service."""

import logging
from typing import Any

import aiohttp
import yaml
from fastapi import status

from ohc_backend.models.ha_entity import Automation, HAEntity, Script

logger = logging.getLogger(__name__)


class HomeAssistantError(Exception):
    """Base exception for Home Assistant API errors."""

    def __init__(self, message: str, status_code: int = None, error_type: str = "ha_error") -> None:
        """Initialize the error."""
        self.message = message
        self.status_code = status_code
        self.error_type = error_type
        super().__init__(message)


class HomeAssistantAuthError(HomeAssistantError):
    """Authentication error with Home Assistant."""

    def __init__(self, message: str = "Authentication with Home Assistant failed") -> None:
        """Initialize the error."""
        super().__init__(message, status_code=401, error_type="ha_auth_error")


class HomeAssistantConnectionError(HomeAssistantError):
    """Connection error with Home Assistant."""

    def __init__(self, message: str = "Could not connect to Home Assistant") -> None:
        """Initialize the error."""
        super().__init__(message, status_code=503, error_type="ha_connection_error")


class HomeAssistantResourceNotFoundError(HomeAssistantError):
    """Resource not found in Home Assistant."""

    def __init__(self, resource_type: str, resource_id: str) -> None:
        """Initialize the error."""
        message = f"{resource_type} not found: {resource_id}"
        super().__init__(message, status_code=404, error_type="ha_not_found")
        self.resource_type = resource_type
        self.resource_id = resource_id


class HomeAssistantService:
    """Service to interact with Home Assistant."""

    def __init__(self, server_url: str, access_token: str, session: aiohttp.ClientSession | None = None,
                 timeout: aiohttp.ClientTimeout | None = None) -> None:
        """Initialize the service."""
        self.server_url = server_url
        self._base_url = f"{server_url}/api"
        self.timeout = timeout or aiohttp.ClientTimeout(total=15)
        self.session = session or aiohttp.ClientSession(
            headers={"Authorization": f"Bearer {access_token}"},
            timeout=self.timeout)

        logger.info(
            "Home Assistant Service initialized with server_url: %s", server_url)

    async def close(self) -> None:
        """Close the service."""
        if self.session and not self.session.closed:
            await self.session.close()
            self.session = None

    async def make_request(self, method: str, url: str, **kwargs: dict) -> Any:
        """Make an HTTP request and return the JSON response with improved error handling."""
        logger.debug("Making %s request to: %s", method, url)

        # Extract resource info from URL for better error messages
        path_parts = url.split("/")
        resource_type = path_parts[-2] if len(path_parts) >= 2 else "resource"
        resource_id = path_parts[-1] if path_parts else "unknown"

        try:
            async with self.session.request(method, url, **kwargs) as response:
                if response.status == status.HTTP_404_NOT_FOUND:
                    logger.debug("%s not found: %s",
                                 resource_type, resource_id)
                    return None

                if response.status == status.HTTP_401_UNAUTHORIZED:
                    error_text = await response.text()
                    logger.error(
                        "Home Assistant authentication error: %s", error_text)
                    msg = f"Authentication failed: {error_text}"
                    raise HomeAssistantAuthError(  # noqa: TRY301
                        msg)

                try:
                    result = await response.json()
                except aiohttp.ContentTypeError as err:
                    content = await response.text()
                    if not response.ok:
                        logger.exception("Invalid JSON response from HA API: %s %s",
                                         response.status, content[:200])
                        msg = f"Invalid JSON response from Home Assistant (Status: {response.status})"
                        raise HomeAssistantError(
                            msg,
                            status_code=response.status
                        ) from err
                    return content  # Return text content if not JSON

                if not response.ok:
                    error_msg = result.get("message", str(result)) if isinstance(
                        result, dict) else str(result)
                    logger.error("Home Assistant API error: %s %s",
                                 response.status, error_msg)
                    msg = f"Home Assistant API error: {error_msg}"
                    raise HomeAssistantError(  # noqa: TRY301
                        msg,
                        status_code=response.status
                    )

                return result
        except aiohttp.ClientConnectorError as e:
            logger.error("Cannot connect to Home Assistant: %s", str(e))
            raise HomeAssistantConnectionError(
                f"Cannot connect to Home Assistant: {e!s}")
        except aiohttp.ClientError as e:
            logger.error("Home Assistant request failed: %s", str(e))
            raise HomeAssistantError(
                f"Request failed: {e!s}", error_type="request_failed")
        except HomeAssistantError:
            raise
        except Exception as e:
            logger.exception("Unexpected error in Home Assistant request")
            raise HomeAssistantError(f"Unexpected error: {e!s}")

    def json_to_yaml(self, content: dict) -> str:
        """Convert JSON content to YAML."""
        try:
            # Convert the dict to YAML string
            return yaml.dump(
                content, default_flow_style=False, sort_keys=False)
        except Exception:
            logger.exception("Error converting json to YAML!")
            raise

    async def get_automation_content(self, automation_id: str) -> str:
        """Get the content of an automation from Home Assistant."""
        json_content = await self.make_request("GET", f"{self._base_url}/config/automation/config/{automation_id}")
        return self.json_to_yaml(json_content)

    async def get_script_content(self, entity_id: str) -> str:
        """Get the content of a script from Home Assistant."""
        name = entity_id.split(".")[1]
        json_content = await self.make_request("GET", f"{self._base_url}/config/script/config/{name}")
        return self.json_to_yaml(json_content)

    async def get_automation(self, automation_id: str) -> Automation:
        """Get a single automation from Home Assistant."""
        return Automation.from_ha_state(await self.make_request("GET",
                                                                f"{self._base_url}/states/automation.{automation_id}"))

    async def get_script(self, script_id: str) -> Script:
        """Get a single automation from Home Assistant."""
        return Script.from_ha_state(await self.make_request("GET", f"{self._base_url}/states/script.{script_id}"))

    async def get_all_automations_and_scripts(self) -> list[HAEntity]:
        """Get both automations and scripts from Home Assistant."""
        states = await self.make_request("GET", f"{self._base_url}/states")

        entities = []

        for state in states:
            entity_id = state["entity_id"]
            if entity_id.startswith(("automation.", "script.")):
                entities.append(HAEntity.from_ha_state(state))
        return entities



"""Home Assistant entity models."""

from __future__ import annotations

from datetime import datetime  # noqa: TC003
from enum import Enum

from pydantic import BaseModel


class HAEntityType(Enum):
    """Home Assistant entity type."""

    AUTOMATION = "automation"
    SCRIPT = "script"


class HAEntity(BaseModel):
    """Base class for Home Assistant entities."""

    entity_id: str
    state: str
    friendly_name: str
    last_changed: datetime
    is_deleted: bool = False

    @classmethod
    def from_ha_state(cls, data: dict) -> HAEntity:
        """Create an entity from Home Assistant state."""
        entity_id = data["entity_id"]
        if entity_id.startswith("automation."):
            return Automation(
                automation_id=data["attributes"]["id"],
                entity_id=entity_id,
                state=data["state"],
                friendly_name=data["attributes"]["friendly_name"],
                last_changed=data["last_changed"],
            )
        if entity_id.startswith("script."):
            return Script(
                entity_id=entity_id,
                state=data["state"],
                friendly_name=data["attributes"]["friendly_name"],
                last_changed=data["last_changed"],
            )
        msg = f"Invalid entity_id: {entity_id}. Must start with 'automation.' or 'script.'"
        raise ValueError(msg)


class Automation(HAEntity):
    """Automation entity."""

    automation_id: str

    @property
    def entity_type(self) -> HAEntityType:
        """Return the entity type."""
        return HAEntityType.AUTOMATION


class Script(HAEntity):
    """Script entity."""

    @property
    def entity_type(self) -> HAEntityType:
        """Return the entity type."""
        return HAEntityType.SCRIPT



import logging
import os
import sys


class CompactExceptionFormatter(logging.Formatter):
    """A formatter that makes exception traces more compact."""

    def formatException(self, exc_info) -> str:
        """Format exception with just the main info, not the full trace."""
        if os.getenv("ENVIRONMENT") == "dev" and os.getenv("DETAILED_LOGS", "").lower() == "true":
            # In development with detailed logs enabled, show full traceback
            return super().formatException(exc_info)
        # Otherwise show a compact version
        exc_type, exc_value, _ = exc_info
        return f"{exc_type.__name__}: {exc_value}"


class RequestIDLogFilter(logging.Filter):
    """Add request ID to log records."""

    def filter(self, record) -> bool:
        """Add request ID to log records."""
        from ohc_backend.utils.request_context import get_request_id
        record.request_id = get_request_id() or "-"
        return True


def configure_logging() -> None:
    """Configure application logging."""
    # Map string log levels to logging constants
    log_level_map = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR
    }

    # Get log level from environment variable, default to INFO if not set or invalid
    log_level_str = os.getenv("LOG_LEVEL", "info").lower()
    log_level = log_level_map.get(log_level_str, logging.INFO)

    # Create formatter with request ID
    log_format = "%(asctime)s [%(levelname)8s] [%(request_id)s] %(name)s: %(message)s"
    formatter = CompactExceptionFormatter(log_format)

    # Configure root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    # Clear existing handlers and add our own
    root_logger.handlers = []

    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)

    # Add the request ID filter to the handler
    request_id_filter = RequestIDLogFilter()
    console_handler.addFilter(request_id_filter)

    root_logger.addHandler(console_handler)

    # Reduce noise from third-party libraries
    logging.getLogger("uvicorn.access").setLevel(logging.WARNING)
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("aiohttp.client").setLevel(logging.WARNING)


def log_error(logger: logging.Logger, message: str, exc: Exception, *, critical: bool = False) -> None:
    """Log an exception with compact formatting."""
    log_method = logger.critical if critical else logger.error

    # Create a clean message with the exception details
    error_message = f"{message}: {type(exc).__name__}: {exc!s}"
    log_method(error_message)

    # If in dev mode with detailed logs, also log full traceback at debug level
    if os.getenv("ENVIRONMENT") == "dev" and os.getenv("DETAILED_LOGS", "").lower() == "true":
        logger.debug("Exception details:", exc_info=exc)



"""Module for handling the GitHub API requests."""

from .client import GitHubClient
from .models import CommitFilesRequest, GithubRepositoryRequestConfig

__all__ = ["CommitFilesRequest", "GitHubClient", "GithubRepositoryRequestConfig"]



"""Pydantic models for GitHub service."""

from pydantic import BaseModel


class TokenResponse(BaseModel):
    """Token response from GitHub."""

    success: bool
    access_token: str | None = None


class DeviceFlowInfo(BaseModel):
    """Device flow authentication information."""

    device_code: str
    user_code: str
    verification_uri: str
    expires_in: int
    interval: int


class Repository(BaseModel):
    """GitHub repository information."""

    full_name: str
    html_url: str
    description: str | None = None
    private: bool = False


class GitBlob(BaseModel):
    """Git blob object."""

    sha: str
    url: str


class GitTreeItem(BaseModel):
    """Git tree item."""

    path: str
    mode: str = "100644"
    type: str = "blob"
    sha: str


class GitTree(BaseModel):
    """Git tree object."""

    sha: str
    url: str
    tree: list[GitTreeItem]


class GitCommit(BaseModel):
    """Git commit object."""

    sha: str
    url: str
    message: str


class CommitFilesRequest(BaseModel):
    """Request to commit multiple files."""

    files: dict[str, str]
    message: str
    branch: str = "main"
    update_only: bool = True


class GithubRepositoryRequestConfig(BaseModel):
    """Configuration for repository creation."""

    name: str
    full_name: str
    private: bool = True
    description: str | None = None


class GitReference(BaseModel):
    """Git reference representing a branch or tag pointer."""

    ref: str
    node_id: str
    url: str
    object: dict



"""HAEntityStorage class, responsible for storing and managing HAEntity objects."""

from __future__ import annotations

import json

from ohc_backend.models.ha_entity import Automation, HAEntity, Script


class OHCState:
    """Storage for HAEntity objects."""

    def __init__(self) -> None:
        """Initialize the storage."""
        self._storage: dict[str, HAEntity] = {}

    def add(self, entity: HAEntity) -> HAEntity:
        """Store a new entity."""
        self._storage[entity.entity_id] = entity
        return entity

    def get_entity(self, entity_id: str) -> HAEntity | None:
        """Get an entity by ID."""
        return self._storage.get(entity_id)

    def get_automation(self, entity_id: str) -> Automation | None:
        """Get an automation by ID."""
        entity = self._storage.get(entity_id)
        if isinstance(entity, Automation):
            return entity
        return None

    def get_script(self, entity_id: str) -> Script | None:
        """Get a script by ID."""
        entity = self._storage.get(entity_id)
        if isinstance(entity, Script):
            return entity
        return None

    def get_automations(self) -> list[Automation]:
        """Get all automations."""
        return [entity for entity in self._storage.values() if isinstance(entity, Automation)]

    def get_scripts(self) -> list[Script]:
        """Get all scripts."""
        return [entity for entity in self._storage.values() if isinstance(entity, Script)]

    def get_entities(self) -> list[HAEntity]:
        """Get all entities."""
        return list(self._storage.values())

    def update(self, entity: HAEntity) -> HAEntity | None:
        """Update an entity."""
        if entity.entity_id in self._storage:
            self._storage[entity.entity_id] = entity
            return entity
        return None

    def remove(self, entity_id: str) -> bool:
        """Delete an entity."""
        if entity_id in self._storage:
            del self._storage[entity_id]
            return True
        return False

    def upsert(self, entity: HAEntity) -> HAEntity:
        """Insert or update an entity."""
        self._storage[entity.entity_id] = entity
        return entity

    def to_json(self) -> str:
        """Serialize the storage to JSON string with pretty printing."""
        return json.dumps(
            self._storage,
            default=lambda x: json.loads(x.model_dump_json()),
            indent=2,
            sort_keys=True,
        )

    @classmethod
    def from_json(cls, json_str: str) -> OHCState:
        """Create a new storage instance from JSON string."""
        storage = cls()
        data = json.loads(json_str)

        type_mapping = {
            "automation": Automation,
            "script": Script,
        }

        for entity_id, entity_data in data.items():
            # Get the entity type from the entity_id prefix
            entity_type = entity_id.split(".")[0]
            entity_class = type_mapping.get(entity_type)

            if entity_class is None:
                msg = f"Unknown entity type: {entity_type}"
                raise ValueError(msg)

            storage._storage[entity_id] = entity_class.model_validate(  # noqa: SLF001
                entity_data)

        return storage



"""Base class for GitHub API communications."""

import logging
from typing import Any

import aiohttp
from fastapi import status

from .errors import (
    GitHubAPIError,
    GitHubAuthError,
    GitHubError,
    GitHubNotFoundError,
    GitHubRateLimitError,
    GitHubValidationError,
)

logger = logging.getLogger(__name__)


class GitHubBaseAPI:
    """Base class for HTTP communications."""

    async def make_request(self, method: str, url: str, **kwargs: dict) -> Any:  # noqa: ANN401
        """Make HTTP request with comprehensive error handling."""
        logger.debug("Making %s request to: %s", method, url)
        try:
            async with self.session.request(method, url, **kwargs) as response:
                # Handle successful responses
                if response.status in (status.HTTP_200_OK, status.HTTP_201_CREATED):
                    return await response.json()

                # Try to parse error response
                error_data = await self._parse_error_response(response)

                # Handle error response outside the try block
                return await self._handle_error_response(response, error_data, url)

        except aiohttp.ClientError as e:
            logger.exception("Network request failed")
            raise GitHubError from e
        except GitHubError:
            # Re-raise GitHub-specific errors
            raise
        except Exception as e:
            # Catch all other exceptions
            logger.exception("Unexpected error in GitHub API request")
            raise GitHubError from e

    async def _parse_error_response(self, response: aiohttp.ClientResponse) -> dict:
        """Parse error response, handling JSON parsing errors."""
        error_data = {}
        try:
            error_data = await response.json()
        except:  # noqa: E722
            error_data = {"message": await response.text() or "Unknown error"}
        return error_data

    async def _handle_error_response(self, response: aiohttp.ClientResponse, error_data: dict, url: str) -> None:
        """Handle different error responses based on status code with improved context."""
        import sys
        from datetime import datetime

        # Capture request context
        method = response.method
        status_code = response.status
        path_parts = url.split("/")
        resource_type = path_parts[-2] if len(path_parts) >= 2 else "resource"
        resource_id = path_parts[-1] if path_parts else "unknown"

        try:
            if status_code == status.HTTP_404_NOT_FOUND:
                raise GitHubNotFoundError(resource_type, resource_id)

            if status_code == status.HTTP_403_FORBIDDEN and "X-RateLimit-Remaining" in response.headers:
                reset_time = int(response.headers.get("X-RateLimit-Reset", 0))
                reset_time_str = datetime.fromtimestamp(
                    reset_time).strftime("%Y-%m-%d %H:%M:%S UTC")
                logger.warning(
                    "GitHub rate limit exceeded. Resets at %s (%d)",
                    reset_time_str, reset_time
                )
                raise GitHubRateLimitError(reset_time)

            if status_code == status.HTTP_401_UNAUTHORIZED:
                # Auth error handling with token cleanup
                from ohc_backend.dependencies import deps
                try:
                    logger.warning(
                        "GitHub authentication failed. Clearing invalid token.")
                    await deps.clear_github_token()
                except Exception as e:
                    logger.exception(
                        "Failed to clear GitHub token: %s", str(e))

                error_msg = error_data.get("message", "Authentication failed")
                raise GitHubAuthError(error_msg)

            if status_code == status.HTTP_422_UNPROCESSABLE_ENTITY:
                errors = error_data.get("errors", [])
                error_details = "; ".join(
                    f"{e.get('field')}: {e.get('message')}" for e in errors) if errors else "Validation failed"
                raise GitHubValidationError(
                    error_data.get("message", "Validation failed") +
                    f" - {error_details}",
                    status_code=422,
                    response_data=error_data
                )

            # Better generic error message
            context = f"{method} {url}"
            error_message = error_data.get(
                "message", f"GitHub API error ({context}): Status {status_code}")

            # Add documentation URL if provided by GitHub
            if "documentation_url" in error_data:
                error_message += f" - See: {error_data['documentation_url']}"

            raise GitHubAPIError(
                error_message,
                status_code=status_code,
                response_data=error_data
            )
        except GitHubError:
            # Add request context to all GitHub errors
            e = sys.exc_info()[1]
            if hasattr(e, "__dict__"):  # Add context attributes to the exception
                e.__dict__["request_url"] = url
                e.__dict__["request_method"] = method
                e.__dict__["status_code"] = status_code
            raise



"""Entry point for running the application."""

import os

import uvicorn

if __name__ == "__main__":
    # Determine if we're in development mode
    debug_mode = os.getenv("ENVIRONMENT", "").lower() == "dev"

    uvicorn.run(
        "ohc_backend.main:app",
        host="0.0.0.0",  # noqa: S104
        port=8000,
        reload=debug_mode
    )



"""Sync manager for Home Assistant entities."""

import asyncio
import copy
import logging
from enum import Enum
from typing import cast

from ohc_backend.models.ha_entity import Automation, HAEntity, HAEntityType
from ohc_backend.services.github import GitHubClient
from ohc_backend.services.github.errors import GitHubAPIError, GitHubAuthError, GitHubNotFoundError
from ohc_backend.services.ha_service import HomeAssistantError, HomeAssistantService
from ohc_backend.services.ohc_state import OHCState
from ohc_backend.services.settings import SyncManagerConfig
from ohc_backend.utils.logging import log_error

logger = logging.getLogger(__name__)


class SyncStatus(Enum):
    """Sync manager status."""

    STOPPED = "stopped"
    RUNNING = "running"
    ERROR = "error"


class SyncManager:
    """Sync manager for Home Assistant entities."""

    def __init__(self, ha_service: HomeAssistantService, github: GitHubClient, sync_config: SyncManagerConfig) -> None:
        """Initialize the sync manager."""
        self.ha_service = ha_service
        self.github = github
        self.sync_config = sync_config

        self._ohc_state: OHCState = OHCState()
        self.status = SyncStatus.STOPPED
        self._task: asyncio.Task | None = None
        self._sleep_task: asyncio.Task | None = None

    def get_ohc_state(self) -> OHCState:
        """Get the ohc state manager."""
        if self.status == SyncStatus.ERROR:
            logger.error("Sync manager is in error state, cannot access state")
            msg = "Sync manager is in error state"
            raise RuntimeError(msg)
        return self._ohc_state

    async def start(self) -> None:
        """Start the sync manager in a background task."""
        if self.status != SyncStatus.STOPPED:
            logger.warning("Sync manager already running or in error state")
            return

        logger.info("Starting sync manager with sync interval %d seconds",
                    self.sync_config.interval)
        try:
            # First load state from GitHub if available
            try:
                logger.info("Loading state from GitHub: %s",
                            self.sync_config.state_file)
                state_json = await self.github.content.get_file_contents(self.sync_config.state_file)
                if state_json:
                    logger.info(
                        "Loading existing state from GitHub (length: %d)", len(state_json))
                    self._ohc_state = OHCState.from_json(state_json)
                    logger.info("State loaded with %d entities",
                                len(self._ohc_state.get_entities()))
                else:
                    logger.warning("State file exists but is empty")
            except GitHubNotFoundError:
                # This is expected the first time when a repo is created.
                logger.info(
                    "State file not found on GitHub, starting with empty state")
                # Already using empty state from initialization

            self.status = SyncStatus.RUNNING
            logger.info("Creating sync task and starting async loop")
            self._task = asyncio.create_task(self._async_loop())
            logger.info("Sync task created")
        except Exception as e:
            log_error(logger, "Failed to start sync manager", e)
            self.status = SyncStatus.ERROR
            raise RuntimeError(f"Failed to start sync manager: {e!s}") from e

    async def stop(self) -> None:
        """Stop the sync manager and cancel any ongoing sleep."""
        logger.info("Stopping sync manager")
        self.status = SyncStatus.STOPPED
        if self._sleep_task and not self._sleep_task.done():
            self._sleep_task.cancel()
        if self._task and not self._task.done():
            await self._task

    async def _async_loop(self) -> None:
        """Run the async loop in background."""
        try:
            while self.status == SyncStatus.RUNNING:
                try:
                    await self.run()

                    # Sleep until next sync
                    try:
                        self._sleep_task = asyncio.create_task(
                            asyncio.sleep(self.sync_config.interval))
                        await self._sleep_task
                    except asyncio.CancelledError:
                        logger.info("Sleep interrupted, stopping sync loop")
                        break
                except Exception as e:
                    log_error(logger, "Error during sync process", e)
                    # Continue running despite errors in a single sync
        except Exception as e:
            log_error(logger, "Fatal error in sync loop", e)
            self.status = SyncStatus.ERROR

    async def run(self) -> None:
        """Run the sync process with improved error handling."""
        try:
            logger.info("Start syncing changes...")

            # Phase 1: Fetch entities and identify changes
            logger.debug("Phase 1: Fetching entities from Home Assistant")
            result = await self._fetch_entities_and_prepare_state()
            if result is None:
                logger.warning("Failed to fetch entities, aborting sync")
                return  # Error occurred during fetch

            ha_entities, state_copy = result
            logger.debug("Phase 1 complete: Retrieved %d entities",
                         len(ha_entities))

            # Phase 2: Process changes and prepare files
            logger.debug("Phase 2: Processing changes")
            result = await self._process_changes(ha_entities, state_copy)
            if result is None:
                logger.warning("Failed to process changes, aborting sync")
                return  # Error occurred during processing

            files, updated, inserted, deleted = result

            # If nothing changed, we're done
            if not (updated or inserted or deleted):
                logger.info("No entities changed, skipping GitHub commit")
                # Still update the internal state to capture last_changed changes
                logger.debug("Updating internal state with %d entities (no content changes)",
                             len(state_copy.get_entities()))
                self._ohc_state = state_copy
                return

            logger.info(
                "Processed entity changes: %d updated, %d inserted, %d deleted",
                len(updated), len(inserted), len(deleted)
            )

            # Phase 3: Commit changes to GitHub
            logger.debug("Phase 3: Committing changes to GitHub")
            success = await self._commit_changes(
                files,
                len(updated),
                len(inserted),
                len(deleted)
            )

            if success:
                # Update the original state only on successful commit
                logger.debug("Updating internal state with %d entities",
                             len(state_copy.get_entities()))
                self._ohc_state = state_copy
                logger.info("Sync completed successfully")

        except Exception as e:
            log_error(logger, "Unexpected error in sync process", e)
            self.status = SyncStatus.ERROR

    async def _fetch_entities_and_prepare_state(self) -> tuple[list[HAEntity], OHCState] | None:
        """Fetch entities from Home Assistant and prepare state copy."""
        try:
            ha_entities = await self.ha_service.get_all_automations_and_scripts()
            logger.debug("Retrieved %d entities from Home Assistant",
                         len(ha_entities))

            # Create a deep copy of ohc_state for modifications
            state_copy = OHCState()
            for entity in self._ohc_state.get_entities():
                state_copy.upsert(copy.deepcopy(entity))

            logger.debug("Created state copy with %d entities",
                         len(state_copy.get_entities()))
            return ha_entities, state_copy

        except HomeAssistantError as e:
            log_error(logger, "Failed to fetch entities from Home Assistant", e)
            return None

    async def fetch_entity_contents_parallel(self, entities: list[HAEntity]) -> list[tuple[HAEntity, str]]:
        """Fetch content for multiple entities in parallel with controlled concurrency."""
        if not entities:
            logger.debug("No entities to fetch content for")
            return []

        logger.debug(
            "Fetching content for %d entities in parallel", len(entities))

        # Create a semaphore to limit concurrency
        semaphore = asyncio.Semaphore(
            self.sync_config.ha_max_parallel_requests)

        async def fetch_with_limit(entity: HAEntity) -> tuple[HAEntity, str]:
            async with semaphore:
                # This ensures only limited coroutines can execute this block at once
                content = await self._fetch_entity_content(entity)
                return entity, content

        # Create tasks for all entities but they'll be limited by the semaphore
        tasks = [fetch_with_limit(entity) for entity in entities]

        try:
            # If any task fails, this will raise an exception
            results = await asyncio.gather(*tasks)
        except Exception as e:
            log_error(
                logger, "Failed to fetch content for some entities, aborting sync", e)
            # Re-raise to abort the current sync
            raise RuntimeError("Entity content fetch failed") from e
        else:
            logger.debug(
                "Successfully fetched content for %d entities", len(results))
            return results

    async def _fetch_entity_content(self, entity: HAEntity) -> str | None:
        """Fetch content for a single entity."""
        try:
            if entity.entity_type == HAEntityType.AUTOMATION:
                automation = cast(Automation, entity)
                return await self.ha_service.get_automation_content(automation.automation_id)
            return await self.ha_service.get_script_content(entity.entity_id)
        except Exception as e:
            log_error(
                logger, f"Failed to fetch content for {entity.entity_id}", e)
            return None

    async def _process_changes(
        self,
        ha_entities: list[HAEntity],
        state_copy: OHCState
    ) -> tuple[dict[str, str], list[HAEntity], list[HAEntity], list[HAEntity]] | None:
        """Process entity changes and prepare files for commit."""
        try:
            # Step 1: Identify all potential changes
            potential_changes = await self._identify_potential_changes(ha_entities, state_copy)
            if potential_changes is None:
                return None

            changed_entities, inserted_entities, deleted_entities = potential_changes

            # Step 2: Check content changes and filter timestamp-only updates
            content_changes = await self._check_content_changes(
                changed_entities, inserted_entities, state_copy)
            if content_changes is None:
                return None

            files, final_changed_entities = content_changes

            # Step 3: Prepare final result
            if final_changed_entities or inserted_entities or deleted_entities:
                # Include state file in commit
                files[self.sync_config.state_file] = state_copy.to_json()
                logger.debug(
                    "Prepared %d files for commit including state file", len(files))
            else:
                logger.debug("No changes detected, no files to commit")

            return files, final_changed_entities, inserted_entities, deleted_entities

        except Exception as e:
            log_error(logger, "Error processing entity changes", e)
            return None

    async def _identify_potential_changes(
        self,
        ha_entities: list[HAEntity],
        state_copy: OHCState
    ) -> tuple[list[HAEntity], list[HAEntity], list[HAEntity]] | None:
        """Identify all potential changes (metadata and timestamp changes)."""
        try:
            changed_entities = []
            inserted_entities = []
            deleted_entities = []

            # Find changed and new entities
            for entity in ha_entities:
                current_entity = state_copy.get_entity(entity.entity_id)
                if not current_entity:
                    # New entity
                    logger.debug("New entity detected: %s", entity.entity_id)
                    state_copy.add(entity)
                    inserted_entities.append(entity)
                else:
                    # Check for any changes including last_changed
                    name_changed = entity.friendly_name != current_entity.friendly_name
                    state_changed = entity.state != current_entity.state
                    timestamp_changed = entity.last_changed != current_entity.last_changed
                    deletion_changed = entity.is_deleted != current_entity.is_deleted
                    was_deleted = current_entity.is_deleted

                    if name_changed or state_changed or timestamp_changed or deletion_changed or was_deleted:
                        if was_deleted:
                            entity.is_deleted = False

                        # Log all changes for debugging
                        if name_changed:
                            logger.debug("Name changed for %s: '%s' -> '%s'",
                                         entity.entity_id, current_entity.friendly_name, entity.friendly_name)
                        if state_changed:
                            logger.debug("State changed for %s: '%s' -> '%s'",
                                         entity.entity_id, current_entity.state, entity.state)
                        if timestamp_changed:
                            logger.debug(
                                "Last changed datetime changed for %s", entity.entity_id)
                        if deletion_changed or was_deleted:
                            logger.debug(
                                "Deletion status changed for %s", entity.entity_id)

                        state_copy.update(entity)
                        changed_entities.append(entity)

            # Find deleted entities
            for entity in state_copy.get_entities():
                if (entity.entity_id not in [e.entity_id for e in ha_entities] and
                        not entity.is_deleted):
                    logger.debug("Entity deleted: %s", entity.entity_id)
                    entity.is_deleted = True
                    state_copy.update(entity)
                    deleted_entities.append(entity)

            logger.debug(
                "Potential changes: %d changed, %d new, %d deleted",
                len(changed_entities), len(
                    inserted_entities), len(deleted_entities)
            )

            return changed_entities, inserted_entities, deleted_entities

        except Exception as e:
            log_error(logger, "Error identifying potential changes", e)
            return None

    async def _check_content_changes(
        self,
        changed_entities: list[HAEntity],
        inserted_entities: list[HAEntity],
        state_copy: OHCState
    ) -> tuple[dict[str, str], list[HAEntity]] | None:
        """Check content changes using SHA comparison."""
        try:
            # Prepare a map of file paths to content
            files_map = {}
            entity_map = {}  # Map of file paths to entities
            path_to_entity_id = {}  # Map file paths to entity_ids for lookups

            # Process all changed and new entities
            entities_to_check = changed_entities + inserted_entities

            if not entities_to_check:
                logger.debug("No entities to check for content changes")
                return {}, []

            # Fetch content for all potentially changed entities
            logger.debug("Fetching content for %d entities",
                         len(entities_to_check))
            content_results = await self.fetch_entity_contents_parallel(entities_to_check)

            # Build file map and entity map
            for entity, content in content_results:
                if not content:
                    logger.warning(
                        "Could not get content for %s, skipping", entity.entity_id)
                    continue

                # Get file path
                prefix = "automations" if entity.entity_type == HAEntityType.AUTOMATION else "scripts"
                file_path = f"{prefix}/{entity.entity_id}.yaml"

                files_map[file_path] = content
                entity_map[file_path] = entity
                path_to_entity_id[file_path] = entity.entity_id

            if not files_map:
                logger.debug("No valid content to check")
                return {}, []

            # Use GitHub client's get_changed_files to efficiently determine changes
            changed_files, _, _ = await self.github.content.get_changed_files(files_map)

            # Process results - entities with content changes
            final_changed_entities = []

            for file_path, content in changed_files.items():
                entity = entity_map.get(file_path)
                if entity:
                    if entity in changed_entities:
                        final_changed_entities.append(entity)
                    # If entity in inserted_entities, it's already handled

            # IMPORTANT: Update state for all entities, even those without content changes
            for entity in entities_to_check:
                # Always update the state in state_copy
                state_copy.update(entity)

            logger.info(
                "SHA-based content analysis: %d entities have actual changes",
                len(final_changed_entities)
            )

            # For entities with no content changes, check if metadata changed (not just timestamp)
            for file_path, entity in entity_map.items():
                if file_path not in changed_files and entity in changed_entities:
                    # Content didn't change, but check if metadata did
                    previous_entity = self._ohc_state.get_entity(
                        entity.entity_id)

                    if previous_entity and (
                        entity.friendly_name != previous_entity.friendly_name or
                        entity.state != previous_entity.state or
                        entity.is_deleted != previous_entity.is_deleted
                    ):
                        # Real metadata changes - include in the commit
                        logger.info("Metadata changed (not just timestamp) for %s - adding to commit",
                                    entity.entity_id)
                        changed_files[file_path] = files_map[file_path]
                        final_changed_entities.append(entity)
                    else:
                        # Only timestamp changed - don't include in commit
                        logger.debug(
                            "Only timestamp changed for %s - not adding to commit", entity.entity_id)

            return changed_files, final_changed_entities

        except Exception as e:
            log_error(logger, "Error checking content changes using SHA", e)
            return None

    async def _commit_changes(
        self,
        files: dict[str, str],
        updated_count: int,
        inserted_count: int,
        deleted_count: int
    ) -> bool:
        """Commit changes to GitHub."""
        try:
            commit_result = await self.github.content.commit_changed_files(
                files,
                f"Commit {updated_count} updated, {inserted_count} new, {deleted_count} deleted automations and/or scripts."
            )

            if commit_result:
                logger.info(
                    "Successfully committed entities to GitHub: %s", list(files.keys()))
                return True

        except GitHubAuthError as e:
            log_error(logger, "GitHub authentication error", e)
            self.status = SyncStatus.ERROR
            return False
        except GitHubAPIError as e:
            log_error(logger, "GitHub API error", e)
            # Don't change status for temporary API issues
            return False
        except Exception as e:
            log_error(logger, "Unexpected error committing to GitHub", e)
            self.status = SyncStatus.ERROR
            return False
        else:
            logger.info(
                "No changes to commit (GitHub reported files unchanged)")
            return True



"""GitHub REST API client and repository manager."""


import aiohttp

from .base import GitHubBaseAPI


class GitHubRestAPI(GitHubBaseAPI):
    """API client for GitHub REST API endpoints."""

    def __init__(self, api_url: str, timeout: aiohttp.ClientTimeout | None = None) -> None:
        """Initialize the GitHub API client."""
        self.timeout = timeout or aiohttp.ClientTimeout(total=15)
        self.session = aiohttp.ClientSession(
            headers={
                "Accept": "application/vnd.github.v3+json",
            },
            timeout=self.timeout,
        )
        self.base_url = api_url

    def set_auth_token(self, token: str) -> None:
        """Set the authentication token."""
        self.session.headers["Authorization"] = f"Bearer {token}"



"""Settings manager for the application."""

import json
import logging
import os
from enum import Enum
from pathlib import Path
from types import UnionType
from typing import Any, Union, get_args, get_origin

import aiofiles
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)


class GithubRepositoryRequestConfig(BaseModel):
    """GitHub repository request configuration."""

    name: str = Field(
        description="Repository name without owner",
        default="ohc-ha-versioning",
    )
    private: bool = Field(
        description="Whether the repository is private",
        default=True,
    )
    description: str | None = Field(
        description="Repository description",
        default="Repo for version control by the OurHomeConnected add-on, on home assistant automation and scripts.",
    )


class GithubConfig(BaseModel):
    """GitHub configuration."""

    repo_request: GithubRepositoryRequestConfig = Field(
        default=GithubRepositoryRequestConfig(),
        description="GitHub repository request configuration",
    )
    access_token: str | None = Field(
        default=None,
    )
    gh_user: str | None = Field(
        default=None,
        description="GitHub username",
    )
    api_url: str = Field(
        default="https://api.github.com",
        description="GitHub API URL",
    )
    client_id: str = Field(
        default="Ov23licd0c0KujXwABCn",
        description="GitHub OAuth client ID",
        json_schema_extra={"env": "GH_CLIENT_ID"}
    )
    scope: str = Field(
        default="repo user:email",
        description="OAuth scopes for GitHub authentication",
    )


class SyncManagerConfig(BaseModel):
    """Sync configuration."""

    interval: int = Field(
        default=300,
        description="Interval in seconds for syncing with Home Assistant",
        json_schema_extra={"env": "HA_SYNC_INTERVAL"},
    )
    state_file: str = Field(
        default=".ohcstate/ohc_state.json",
        description="Path to state file",
    )
    ha_max_parallel_requests: int = Field(
        default=10,
        description="Maximum number of parallel requests to Home Assistant",
        json_schema_extra={"env": "HA_MAX_PARALLEL_REQUESTS"},
    )


class HAConfig(BaseModel):
    """Home Assistant configuration."""

    server: str = Field(
        default="http://supervisor.core",
        description="Home Assistant server URL",
        json_schema_extra={"env": "HA_SERVER"},
    )
    token: str | None = Field(
        default=None,
        description="Home Assistant authentication token",
        json_schema_extra={"env": "SUPERVISOR_TOKEN"},
    )
    data_folder: str = Field(
        default="/data",
        description="Path to data folder",
        json_schema_extra={"env": "HA_DATA_FOLDER"},
    )


class AppSettings(BaseModel):
    """Application settings."""

    ha: HAConfig | None = Field(
        default=HAConfig(),
        description="Home Assistant configuration",
    )
    gh: GithubConfig | None = Field(
        default=GithubConfig(),
        description="GitHub configuration",
    )
    sync: SyncManagerConfig = Field(
        default=SyncManagerConfig(),
        description="Sync configuration",
    )


class Settings:
    """Settings manager for the application."""

    def __init__(self, file_path: str) -> None:
        """Initialize the settings manager."""
        self._file_path = Path(file_path)
        self._settings = AppSettings()

    async def async_init(self) -> None:
        """Initialize settings."""
        await self._load()

    async def _load(self) -> None:
        """Load settings from file and environment variables."""
        # Load from file first
        if self._file_path.exists():
            try:
                async with aiofiles.open(self._file_path) as f:
                    content = await f.read()
                    file_settings = json.loads(content)
                    self._settings = AppSettings(**file_settings)
            except (json.JSONDecodeError, ValueError):
                logger.exception("Error loading settings file.")

        # Load from environment variables
        self._load_env_vars(self._settings)

    def _load_env_vars(self, model: BaseModel) -> None:
        """Recursively load environment variables for a model and its nested models."""
        for field_name, field in model.model_fields.items():
            # Get current value
            value = getattr(model, field_name)

            # Check if field has env variable defined
            extra = getattr(field, "json_schema_extra", {}) or {}
            if (env_var := extra.get("env")) and (env_value := os.environ.get(env_var)):
                try:
                    parsed_value = self._parse_value(
                        env_value, field.annotation)
                    setattr(model, field_name, parsed_value)
                except ValueError:
                    logger.exception(
                        "Error parsing environment variable %s", env_var)

            # Handle nested models
            if isinstance(value, BaseModel):
                self._load_env_vars(value)
            elif value is None and issubclass(get_origin(field.annotation) or field.annotation, BaseModel):
                model_class = get_args(field.annotation)[0] if get_origin(
                    field.annotation) else field.annotation
                new_model = model_class()
                self._load_env_vars(new_model)
                setattr(model, field_name, new_model)

    def _parse_value(self, value: str, type_: type) -> Any:  # noqa: ANN401
        """Parse string value to target type."""
        # Handle Union types (e.g., str | None)
        if isinstance(type_, UnionType) or get_origin(type_) is Union:
            types = get_args(type_)
            # Use first non-None type
            actual_type = next(t for t in types if t is not type(None))
            return self._parse_value(value, actual_type)

        if type_ is bool:
            return value.lower() in ("true", "1", "yes")
        if isinstance(type_, type) and issubclass(type_, BaseModel):
            return type_(**json.loads(value))
        if isinstance(type_, type) and issubclass(type_, Enum):
            return type_(value)
        return type_(value)

    async def save(self) -> None:
        """Save current settings to disk."""
        self._file_path.parent.mkdir(parents=True, exist_ok=True)
        async with aiofiles.open(self._file_path, "w") as f:
            await f.write(self._settings.model_dump_json(indent=2))

    async def update(self, **kwargs: dict[str, Any]) -> None:
        """Update settings with new values."""
        new_settings = self._settings.model_copy(update=kwargs)
        # Validate before saving
        self._settings = new_settings
        await self.save()

    def get_env_mappings(self) -> dict[str, str]:
        """Get the mapping of setting names to environment variables."""
        return self._env_mappings.copy()

    def describe(self) -> dict[str, dict[str, Any]]:
        """Get descriptions and metadata for all settings."""
        descriptions = {}
        for field_name, field in AppSettings.model_fields.items():
            descriptions[field_name] = {
                "description": field.description,
                "type": str(field.annotation.__name__),
                "default": field.default,
                "env_var": self._env_mappings.get(field_name, ""),
                "current_value": getattr(self._settings, field_name),
            }
        return descriptions

    # Type hints for common attributes
    @property
    def gh_config(self) -> GithubConfig:
        """Get the GitHub configuration."""
        return self._settings.gh

    @property
    def gh_token(self) -> str:
        """Get the GitHub access token."""
        return self._settings.gh.access_token

    @property
    def ha_config(self) -> HAConfig:
        """Get the Home Assistant configuration."""
        return self._settings.ha

    @property
    def sync_config(self) -> SyncManagerConfig:
        """Get the sync configuration."""
        return self._settings.sync
